// Read/Write request processors for AIOWDM driver

#include "stddcls.h"
#include "driver.h"

NTSTATUS GetClearIOFor(PDEVICE_EXTENSION);

#pragma LOCKEDDATA

PDEVICE_EXTENSION GlobalExtension[MAX_SLOTS_OPEN];

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS DispatchCreate(PDEVICE_OBJECT fdo, PIRP Irp)
	{							// DispatchCreate
	PAGED_CODE();
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;

	PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);

	// Claim the remove lock in Win2K so that removal waits until the
	// handle closes. Don't do this in Win98, however, because this
	// device might be removed by surprise with handles open, whereupon
	// we'll deadlock in HandleRemoveDevice waiting for a close that
	// can never happen because we can't run the user-mode code that
	// would do the close.

	NTSTATUS status;
	if ( win98 || ! pdx->bRealDevice )
		status = STATUS_SUCCESS;
	else
		status = IoAcquireRemoveLock(&pdx->RemoveLock, stack->FileObject);

	if (NT_SUCCESS(status))
	{						// okay to open
		if (InterlockedIncrement(&pdx->handles) == 1)
		{					// first open handle
		}					// first open handle
		KdPrint(("Opened handle, %d handle%s open\n", pdx->handles, pdx->handles-1 ? "s" : ""));
	}					// okay to open
	return CompleteRequest(Irp, status, 0);
	}							// DispatchCreate

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

void ForgetIRP(PIRP IRP)
{
	PAGED_CODE();
	//This function is evil. However, if we complete the IRP, it'll ret into a thread that no
	//longer exists. This way, it'll remain pending indefinitely, consuming resources
	//including nonpaged memory... but the system won't bugcheck.

	//It would be really nice of us to delete the IRP without completing it, but we don't
	//know how, and furthermore don't believe there's a way(since it wouldn't be necessary
	//without this "ret into a thread that no longer exists" bug).
}

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS DispatchClose(PDEVICE_OBJECT fdo, PIRP Irp)
	{							// DispatchClose
	PAGED_CODE();
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;
	PIO_STACK_LOCATION stack = IoGetCurrentIrpStackLocation(Irp);
	if (InterlockedDecrement(&pdx->handles) == 0)
	{						// no more open handles
		if (pdx->bRealDevice && pdx->InterruptRequested && pdx->HasPendingIrp)
		{
			pdx->InterruptRequested = FALSE;
			pdx->HasPendingIrp = FALSE;
			ForgetIRP(pdx->PendingIrp);
			pdx->PendingIrp = NULL;
			pdx->InterruptOccurred = FALSE;
		}
	}						// no more open handles
	KdPrint(("Closed handle, %d handle%s open\n", pdx->handles, pdx->handles-1 ? "s" : ""));

	// Release the remove lock to match the acquisition done in DispatchCreate

	if ( pdx->bRealDevice && ! win98 )
		IoReleaseRemoveLock(&pdx->RemoveLock, stack->FileObject);

	return CompleteRequest(Irp, STATUS_SUCCESS, 0);
	}							// DispatchClose

///////////////////////////////////////////////////////////////////////////////

#pragma LOCKEDCODE

VOID DpcForIsr(PKDPC Dpc, PDEVICE_OBJECT fdo, PIRP junk, PDEVICE_EXTENSION pdx)
{							// TrueDPC
	DoIOAction(pdx, pdx->Enable);

	if (pdx->InterruptRequested && pdx->HasPendingIrp)
	{
		pdx->InterruptRequested = FALSE;
		pdx->InterruptOccurred = FALSE;
		pdx->HasPendingIrp = FALSE;
		pdx->PendingIrp->IoStatus.Status = STATUS_NO_MORE_FILES;//I wish STATUS_SUCCESS worked...
		pdx->PendingIrp->IoStatus.Information = pdx->COSSize;
		IoSetCancelRoutine(pdx->PendingIrp, NULL);
		IoCompleteRequest(pdx->PendingIrp, IO_SOUND_INCREMENT);
	}
	else
	{
		pdx->InterruptOccurred = TRUE;
	}

}							// TrueDPC

///////////////////////////////////////////////////////////////////////////////

#pragma LOCKEDCODE

BOOLEAN TrueISR(PDEVICE_EXTENSION pdx)
{							// TrueISR
	if ( ! pdx ) return FALSE; //Why did it ask if I have noone to answer for?
	if ( ! pdx->bRealDevice ) return FALSE; //I don't represent a device, it wasn't me.

	KdPrint((DRIVERNAME " - IRQ from %04LX?\n", pdx->DeviceID));
	UCHAR Data = QueryIOAction(pdx, pdx->IAm);

	if ( ! Data ) return FALSE; //then it wasn't me

	KdPrint((DRIVERNAME " - Yes!\n"));


	DoIOAction(pdx, pdx->Disable);

	if (pdx->InterruptRequested && pdx->HasPendingIrp)
		InCOSData(pdx, pdx->PendingIrp, pdx->COSSize)
	;

	DoIOAction(pdx, pdx->Clear);
	//The DPC enables
	IoRequestDpc(pdx->DeviceObject, NULL, pdx);


	return TRUE;
}							// TrueISR

BOOLEAN OnInterrupt(PKINTERRUPT InterruptObject, PVOID corrupt)
{
	KdPrint((DRIVERNAME " - IRQ...\n"));
	BOOLEAN Anybody = FALSE;

	for (int I = MAX_SLOTS_OPEN - 1; I >= 0 && GlobalExtension[I]; --I)
		if (TrueISR(GlobalExtension[I]))
			Anybody = TRUE;

	return Anybody;
}

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

// Map port address, if on RISC platform.
AlphaMapSpace(BOOLEAN bMap, PUCHAR * ToStoreMapped, PHYSICAL_ADDRESS Nonmapped, ULONG Size)
{
	if ( ! bMap )
		*ToStoreMapped = (PUCHAR) Nonmapped.QuadPart;
	else
	{
		*ToStoreMapped = (PUCHAR) MmMapIoSpace(Nonmapped, Size, MmNonCached);
		if ( ! *ToStoreMapped )
		{
			KdPrint((DRIVERNAME " - Unable to map port range %I64X, length %X\n", Nonmapped, Size));
			return STATUS_INSUFFICIENT_RESOURCES;
		}
	}
	return STATUS_SUCCESS;
}
///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

NTSTATUS StartDevice(PDEVICE_OBJECT fdo, PCM_PARTIAL_RESOURCE_LIST raw, PCM_PARTIAL_RESOURCE_LIST translated)
{							// StartDevice
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;
	NTSTATUS status;

	// Identify the I/O resources we're supposed to use.

	ULONG vector;
	KIRQL irql;
	KINTERRUPT_MODE mode;
	KAFFINITY affinity;
	BOOLEAN irqshare;
	BOOLEAN gotinterrupt = FALSE;
	//ULONG StatusSize, AuxSize;
	//BOOLEAN gotport = FALSE, GotStatus = FALSE, GotAux = FALSE;
	TBarData FoundBar[6];
	ULONG iFoundBar = 0;
	ULONG FoundBarCt = 0;

	if ( ! translated ) return STATUS_DEVICE_CONFIGURATION_ERROR; //No resources assigned?

	PCM_PARTIAL_RESOURCE_DESCRIPTOR resource = translated->PartialDescriptors;
	ULONG nres = translated->Count;
	for (ULONG i = 0; i < nres; ++i, ++resource)
	{						// for each resource
		switch (resource->Type)
		{					// switch on resource type

			case CmResourceTypePort: case CmResourceTypeMemory:
				FoundBar[iFoundBar++] = BarFromResource(*resource);
			break;

			case CmResourceTypeInterrupt:
				irql = (KIRQL) resource->u.Interrupt.Level;
				vector = resource->u.Interrupt.Vector;
				affinity = resource->u.Interrupt.Affinity;
				mode = (resource->Flags == CM_RESOURCE_INTERRUPT_LATCHED) ?
					Latched : LevelSensitive;
				irqshare = resource->ShareDisposition == CmResourceShareShared;
				gotinterrupt = TRUE;
			break;

			case CmResourceTypeDevicePrivate:
				break;
			default:
				KdPrint((DRIVERNAME " - Unexpected I/O resource type %d\n", resource->Type));
			break;
		}					// switch on resource type
	}						// for each resource
	FoundBarCt = iFoundBar;

	pdx->StatusBar = BlankBarData;
	pdx->MainBar = BlankBarData;
	pdx->AuxBar = BlankBarData;
	switch ( FoundBarCt )
	{
		case 0:
			KdPrint((DRIVERNAME " - Didn't get needed I/O resources\n"));
			return STATUS_DEVICE_CONFIGURATION_ERROR;
		break;
		case 1:
			pdx->MainBar = FoundBar[0];
		break;
		case 2:
			pdx->StatusBar = FoundBar[0];
			pdx->MainBar = FoundBar[1];
		break;
		default: for ( ULONG I = 0; I < FoundBarCt; ++I )
		{
			if ( FoundBar[I].BarType == btMemory )
			{
				DelBar(FoundBar[I]);
			}
			else if ( pdx->MainBar.Base == NULL )
			{
				pdx->MainBar = FoundBar[I];
			}
			else if ( pdx->StatusBar.Base == NULL )
			{
				//The previous was the status bar, this is the main bar.
				pdx->StatusBar = pdx->MainBar;
				pdx->MainBar = FoundBar[I];
			}
			else if ( pdx->AuxBar.Base == NULL )
			{
				pdx->AuxBar = FoundBar[I];
			}
			else
			{
				DelBar(FoundBar[I]);
			}
		}
		break;
	}

	status = GetClearIOFor(pdx);
	if ( ! NT_SUCCESS(status) )
	{
		KdPrint((DRIVERNAME " - Could not get clear I/O methods!\n"));
		DelBar(pdx->StatusBar);
		DelBar(pdx->MainBar);
		DelBar(pdx->AuxBar);
		return STATUS_DEVICE_CONFIGURATION_ERROR;
	}
	if ( pdx->IAm.Operation == 'N' )
	{
		if ( pdx->StatusBar.Base == NULL )
		{
			//If there is no status bar, then this is a subtractive-decode device.
			pdx->IAm = BlankIOAction;
		}
		else if ( pdx->StatusBar.Size >= 0x100 )
		{
			pdx->IAm.BarID = 'S';
			pdx->IAm.Operation = 'R';
			pdx->IAm.Offset = 0x69;
			pdx->IAm.Data = 0x80;
		}
		else
		{
			pdx->IAm.BarID = 'S';
			pdx->IAm.Operation = 'R';
			pdx->IAm.Offset = 0x4C;
			pdx->IAm.Data = 0x04;
		}
	}

	DoIOAction(pdx, pdx->Disable);
	DoIOAction(pdx, pdx->Clear);
	//DoIOAction(pdx, pdx->Enable);

	//Attempts to correct .BIN error. Doesn't work.
	//if ( pdx->StatusBar.Size >= 0x100 ) OutBar32(pdx->StatusBar, 0x34, 0x00000050);

	pdx->GotIRQ = gotinterrupt;
	if ( pdx->GotIRQ )
	{
		status = IoConnectInterrupt(&pdx->InterruptObject, OnInterrupt,
			(PVOID) pdx, NULL, vector, irql, irql, mode, irqshare, affinity, FALSE);
		if (!NT_SUCCESS(status))
		{
			KdPrint((DRIVERNAME " - IoConnectInterrupt failed - %X\n", status));
			DelBar(pdx->StatusBar);
			DelBar(pdx->MainBar);
			DelBar(pdx->AuxBar);
			return status;
		}
		if ( pdx->StatusBar.Size >= 0x100 ) OutBar8(pdx->StatusBar, 0x69, 0x09); // enable IRQs on PCIe bridge chip
	}
	else
	{
		pdx->InterruptObject = NULL;
	}

	KdPrint((DRIVERNAME " - DeviceID = %04X, main base = %LX, status base = %LX\n", pdx->DeviceID, pdx->MainBar.Base, pdx->StatusBar.Base));

	return STATUS_SUCCESS;
}							// StartDevice

///////////////////////////////////////////////////////////////////////////////

#pragma PAGEDCODE

VOID StopDevice(IN PDEVICE_OBJECT fdo, BOOLEAN oktouch /* = FALSE */)
{							// StopDevice
	PDEVICE_EXTENSION pdx = (PDEVICE_EXTENSION) fdo->DeviceExtension;

	if (pdx->InterruptObject)
		{						// disconnect interrupt

		DoIOAction(pdx, pdx->Disable);

		IoDisconnectInterrupt(pdx->InterruptObject);
		pdx->InterruptObject = NULL;
		}						// disconnect interrupt

	DelBar(pdx->StatusBar);
	DelBar(pdx->MainBar);
	DelBar(pdx->AuxBar);
}							// StopDevice

